@@ -1,7 +1,7 @@
 #![feature(hash_drain_filter,async_closure)]
 use std::collections::HashMap;
 use lru::LruCache;
-use async_std::{prelude::*,fs,sync::{RwLock,Arc,Mutex},io};
+use async_std::{prelude::*,fs,sync::{RwLock,Arc,Mutex},task};
 use desert::{varint,ToBytes,FromBytes};
 
 mod storage;
@@ -16,12 +16,12 @@ pub type RecordId = u64;
 pub type IdBlock = u64;
 pub type Error = Box<dyn std::error::Error+Send+Sync>;
 
-pub trait Record: Send+Sync+Clone+std::fmt::Debug {
+pub trait Record: Send+Sync+Clone+std::fmt::Debug+'static {
   fn get_id(&self) -> RecordId;
   fn get_refs(&self) -> Vec<RecordId>;
   fn get_position(&self) -> Option<Position>;
   fn pack(records: &HashMap<RecordId,Self>) -> Vec<u8> where Self: Sized;
-  fn unpack(buf: &[u8]) -> Result<HashMap<RecordId,Self>,Error>;
+  fn unpack(buf: &[u8], records: &mut HashMap<RecordId,Self>) -> Result<usize,Error>;
 }
 
 #[derive(Debug,Clone)]
@@ -40,20 +40,22 @@ impl QTree {
 
 // todo: backrefs
 pub struct XQ<S,R> where S: RW, R: Record {
-  storage: Mutex<Box<dyn Storage<S>>>,
-  active_files: HashMap<String,Arc<Mutex<()>>>,
+  storage: Arc<Mutex<Box<dyn Storage<S>>>>,
+  active_files: Arc<RwLock<HashMap<String,Arc<Mutex<()>>>>>,
   root: QTree,
-  quad_cache: LruCache<QuadId,HashMap<RecordId,R>>,
+  record_cache: LruCache<RecordId,R>,
   quad_updates: Arc<RwLock<HashMap<QuadId,Option<HashMap<RecordId,R>>>>>,
   quad_update_age: HashMap<QuadId,usize>,
-  quad_update_count: u64,
-  quad_count: HashMap<QuadId,u64>,
+  quad_update_count_total: u64,
+  quad_update_count: HashMap<QuadId,u64>, // count of staged updates in memory
+  quad_count: HashMap<QuadId,u64>, // count of staged updates in memory+existing data on disk
+  quad_bbox: HashMap<QuadId,BBox>,
   next_quad_id: QuadId,
-  id_cache: LruCache<IdBlock,HashMap<RecordId,QuadId>>,
+  id_cache: Arc<Mutex<LruCache<RecordId,QuadId>>>,
   id_updates: Arc<RwLock<HashMap<IdBlock,HashMap<RecordId,QuadId>>>>,
-  id_update_age: HashMap<IdBlock,usize>,
-  id_update_count: u64,
-  id_count: HashMap<IdBlock,u64>,
+  id_update_age: HashMap<IdBlock,usize>, // should be called flush age
+  id_update_count_total: u64,
+  id_update_count: HashMap<IdBlock,u64>,
   missing_updates: HashMap<RecordId,R>,
   missing_count: usize,
   next_missing_id: u64,
@@ -67,8 +69,8 @@ pub struct Fields {
   pub id_flush_size: u64,
   pub id_flush_top: usize,
   pub id_flush_max_age: usize,
+  pub record_cache_size: usize,
   pub quad_block_size: u64,
-  pub quad_cache_size: usize,
   pub quad_flush_size: u64,
   pub quad_flush_top: usize,
   pub quad_flush_max_age: usize,
@@ -78,16 +80,16 @@ pub struct Fields {
 impl Default for Fields {
   fn default() -> Self {
     Self {
-      id_block_size: 500_000,
-      id_cache_size: 5_000,
-      id_flush_size: 10_000_000,
+      id_block_size: 50_000,
+      id_cache_size: 100_000_000,
+      id_flush_size: 20_000_000,
       id_flush_top: 200,
       id_flush_max_age: 40,
       quad_block_size: 50_000,
-      quad_cache_size: 2_000,
-      quad_flush_size: 10_000_000,
+      record_cache_size: 5_000_000,
+      quad_flush_size: 5_000_000,
       quad_flush_top: 200,
-      quad_flush_max_age: 20,
+      quad_flush_max_age: 40,
       missing_flush_size: 2_000_000,
     }
   }
@@ -109,23 +111,28 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
       quad_updates.insert(0, None);
       let mut quad_count = HashMap::new();
       quad_count.insert(0, 0);
+      let mut quad_bbox = HashMap::new();
+      quad_bbox.insert(0, (-180.0,-90.0,180.0,90.0));
+
       let xq = Self {
-        storage: Mutex::new(storage),
+        storage: Arc::new(Mutex::new(storage)),
         root: QTree::Quad {
           id: 0,
           bbox: (-180.0,-90.0,180.0,90.0),
         },
-        active_files: HashMap::new(),
-        quad_cache: LruCache::new(fields.quad_cache_size),
+        quad_bbox,
+        active_files: Arc::new(RwLock::new(HashMap::new())),
+        record_cache: LruCache::new(fields.record_cache_size),
         quad_updates: Arc::new(RwLock::new(quad_updates)),
-        quad_update_count: 0,
+        quad_update_count_total: 0,
+        quad_update_count: HashMap::new(),
         quad_update_age: HashMap::new(),
         quad_count,
-        id_cache: LruCache::new(fields.id_cache_size),
+        id_cache: Arc::new(Mutex::new(LruCache::new(fields.id_cache_size))),
         id_updates: Arc::new(RwLock::new(HashMap::new())),
         id_update_age: HashMap::new(),
-        id_update_count: 0,
-        id_count: HashMap::new(),
+        id_update_count_total: 0,
+        id_update_count: HashMap::new(),
         missing_updates: HashMap::new(),
         missing_count: 0,
         next_quad_id: 1,
@@ -136,20 +143,23 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     } else {
       let (_,meta) = Meta::from_bytes(&buf)?;
       fields.id_block_size = meta.id_block_size;
+      let quad_bbox = Self::get_bboxes(&meta.root);
       let xq = Self {
-        storage: Mutex::new(storage),
+        storage: Arc::new(Mutex::new(storage)),
         root: meta.root,
-        active_files: HashMap::new(),
-        quad_cache: LruCache::new(fields.quad_cache_size),
+        quad_bbox,
+        active_files: Arc::new(RwLock::new(HashMap::new())),
+        record_cache: LruCache::new(fields.record_cache_size),
         quad_updates: Arc::new(RwLock::new(HashMap::new())),
-        quad_update_count: 0,
+        quad_update_count_total: 0,
+        quad_update_count: HashMap::new(),
         quad_update_age: HashMap::new(),
         quad_count: meta.quad_count,
-        id_cache: LruCache::new(fields.id_cache_size),
+        id_cache: Arc::new(Mutex::new(LruCache::new(fields.id_cache_size))),
         id_updates: Arc::new(RwLock::new(HashMap::new())),
         id_update_age: HashMap::new(),
-        id_update_count: 0,
-        id_count: HashMap::new(),
+        id_update_count_total: 0,
+        id_update_count: HashMap::new(),
         missing_updates: HashMap::new(),
         missing_count: 0,
         next_quad_id: meta.next_quad_id,
@@ -181,25 +191,49 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     }
     quad_ids
   }
-  pub async fn read_quad(&mut self, q_id: QuadId) -> Result<HashMap<RecordId,R>,Error> {
-    let mut records = {
-      if let Some(records) = self.quad_cache.get(&q_id) {
-        records.clone()
-      } else {
-        let qfile = quad_file(q_id);
-        let records = match self.open_file_r(&qfile).await? {
-          Some(mut s) => {
-            let mut buf = vec![];
-            s.read_to_end(&mut buf).await?;
-            R::unpack(&buf)?
+  fn get_bboxes(root: &QTree) -> HashMap<QuadId,BBox> {
+    let mut cursors = vec![root];
+    let mut ncursors = vec![];
+    let mut bboxes = HashMap::new();
+    while !cursors.is_empty() {
+      ncursors.clear();
+      for c in cursors.iter() {
+        match c {
+          QTree::Node { children, .. } => {
+            ncursors.extend(children)
           },
-          None => HashMap::new(),
-        };
-        self.quad_cache.put(q_id, records.clone());
-        self.close_file(&qfile);
-        records
+          QTree::Quad { id, bbox } => {
+            bboxes.insert(*id, bbox.clone());
+          },
+        }
       }
+      let tmp = ncursors;
+      ncursors = cursors;
+      cursors = tmp;
+    }
+    bboxes
+  }
+  pub async fn read_quad(&mut self, q_id: QuadId) -> Result<HashMap<RecordId,R>,Error> {
+    let qfile = quad_file(q_id);
+    let mut records = match self.open_file_r(&qfile).await? {
+      Some(mut s) => {
+        let mut buf = vec![];
+        s.read_to_end(&mut buf).await?;
+        let mut rs = HashMap::new();
+        let mut offset = 0;
+        while offset < buf.len() {
+          let s = R::unpack(&buf[offset..], &mut rs)?;
+          offset += s;
+          if s == 0 { break }
+        }
+        rs
+      },
+      None => HashMap::new(),
     };
+    self.close_file(&qfile).await;
+    for (id,r) in records.iter() {
+      self.record_cache.put(*id, r.clone());
+    }
     if let Some(Some(xrecords)) = self.quad_updates.read().await.get(&q_id) {
       for (id,r) in xrecords {
         records.insert(*id,r.clone());
@@ -246,11 +280,11 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     let b = self.id_block(id);
     if let Some(ids) = self.id_updates.write().await.get_mut(&b) {
       ids.insert(id, q_id);
-      self.id_update_count += 1;
-      if let Some(c) = self.id_count.get_mut(&b) {
+      self.id_update_count_total += 1;
+      if let Some(c) = self.id_update_count.get_mut(&b) {
         *c += 1;
       } else {
-        self.id_count.insert(b, 1);
+        self.id_update_count.insert(b, 1);
       }
       return Ok(());
     }
@@ -258,11 +292,11 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
       let mut ids = HashMap::new();
       ids.insert(id, q_id);
       self.id_updates.write().await.insert(b, ids);
-      self.id_update_count += 1;
-      if let Some(c) = self.id_count.get_mut(&b) {
+      self.id_update_count_total += 1;
+      if let Some(c) = self.id_update_count.get_mut(&b) {
         *c += 1;
       } else {
-        self.id_count.insert(b, 1);
+        self.id_update_count.insert(b, 1);
       }
     }
     Ok(())
@@ -271,7 +305,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     if records.is_empty() { return Ok(()) }
     let qs = self.get_quads(&records).await?;
     if qs.is_empty() { return Ok(()) }
-    for (q_id,(bbox,ix)) in qs.iter() {
+    for (q_id,ix) in qs.iter() {
       for i in ix {
         let record = records.get(*i).unwrap();
         self.insert_id(record.get_id(), *q_id).await?;
@@ -291,26 +325,37 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
           }
           qu.insert(*q_id, Some(items));
         }
-        self.quad_update_count += ix.len() as u64;
-        if let Some(c) = self.quad_count.get_mut(q_id) {
-          *c += ix.len() as u64;
-        } else {
-          self.quad_count.insert(*q_id, ix.len() as u64);
-        }
       }
-      if self.quad_count.get(q_id).cloned().unwrap_or(0) > self.fields.quad_block_size {
-        self.split_quad(&q_id, &bbox).await?;
+      self.quad_add_update_count(q_id, ix.len() as u64);
+      if self.quad_get_update_count(q_id) > self.fields.quad_block_size {
+        self.split_quad(&q_id).await?;
       }
     }
     self.check_flush().await?;
     Ok(())
   }
+  fn quad_add_update_count(&mut self, q_id: &QuadId, n: u64) {
+    self.quad_update_count_total += n;
+    if let Some(c) = self.quad_count.get_mut(q_id) {
+      *c += n;
+    } else {
+      self.quad_count.insert(*q_id, n);
+    }
+    if let Some(c) = self.quad_update_count.get_mut(q_id) {
+      *c += n;
+    } else {
+      self.quad_update_count.insert(*q_id, n);
+    }
+  }
+  fn quad_get_update_count(&self, q_id: &QuadId) -> u64 {
+    self.quad_count.get(q_id).cloned().unwrap_or(0)
+  }
   pub async fn check_flush(&mut self) -> Result<(),Error> {
     // todo: parallel io
-    if self.quad_update_count >= self.fields.quad_flush_size {
+    if self.quad_update_count_total >= self.fields.quad_flush_size {
       self.quad_flush_partial(self.fields.quad_flush_top).await?;
     }
-    if self.id_update_count >= self.fields.id_flush_size {
+    if self.id_update_count_total >= self.fields.id_flush_size {
       self.id_flush_partial(self.fields.id_flush_top).await?;
     }
     if self.missing_updates.len() as u64 >= self.fields.missing_flush_size {
@@ -327,56 +372,65 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     let rs: HashMap<u64,R> = o_rs.unwrap();
     if rs.is_empty() { return Ok(()) }
     let qfile = quad_file(q_id);
-    let mut s = self.open_file_rw(&qfile).await?;
-    let buf = {
-      if let Some(records) = self.quad_cache.get_mut(&q_id) {
-        for (r_id,r) in rs.iter() {
-          records.insert(*r_id, r.clone());
-        }
-        R::pack(&records)
-      } else {
-        let mut buf = vec![];
-        s.read_to_end(&mut buf).await?;
-        s.seek(io::SeekFrom::Start(0)).await?;
-        let mut records = R::unpack(&buf)?;
-        for (r_id,r) in rs.iter() {
-          records.insert(*r_id, r.clone());
-        }
-        self.quad_cache.put(q_id, records.clone());
-        R::pack(&records)
-      }
-    };
-    s.set_len(buf.len() as u64).await?;
+    let mut s = self.open_file_a(&qfile).await?;
+    let buf = R::pack(&rs);
     s.write_all(&buf).await?;
     s.flush().await?;
-    self.close_file(&qfile);
-    self.quad_update_count -= rs.len() as u64;
-    self.quad_count.insert(q_id, 0);
+    self.close_file(&qfile).await;
+    self.quad_update_count_total -= rs.len() as u64;
+    self.quad_update_count.insert(q_id, 0);
     self.quad_update_age.insert(q_id, 0);
     Ok(())
   }
-  pub async fn quad_flush_all(&mut self) -> Result<(),Error> {
-    // todo: parallel io
-    let qs = self.quad_updates.read().await.keys().copied().collect::<Vec<QuadId>>();
-    for q_id in qs {
-      self.quad_flush_by_id(q_id).await?;
+  pub async fn quad_flush_by_ids(&mut self, q_ids: &[QuadId]) -> Result<(),Error> {
+    let mut work: Vec<task::JoinHandle<
+      Result<(QuadId,u64),Error>
+    >> = Vec::with_capacity(q_ids.len());
+    for q in q_ids {
+      let q_id = *q;
+      let active_files = self.active_files.clone();
+      let storage = self.storage.clone();
+      let quc = self.quad_updates.clone();
+      work.push(task::spawn(async move {
+        let o_rs = {
+          let mut qu = quc.write().await;
+          qu.remove(&q_id).and_then(|v| v)
+        };
+        if o_rs.is_none() { return Ok((q_id, 0)) }
+        let rs: HashMap<u64,R> = o_rs.unwrap();
+        if rs.is_empty() { return Ok((q_id, 0)) }
+        let qfile = quad_file(q_id);
+        let mut s = Self::open_file_a_params(&qfile, active_files.clone(), storage).await?;
+        let buf = R::pack(&rs);
+        s.write_all(&buf).await?;
+        s.flush().await?;
+        Self::close_file_params(&qfile, active_files).await;
+        Ok((q_id, rs.len() as u64))
+      }));
+    }
+    for r in futures::future::join_all(work).await {
+      let (q_id,n) = r?;
+      self.quad_update_count_total -= n;
+      self.quad_update_count.insert(q_id, 0);
+      self.quad_update_age.insert(q_id, 0);
     }
     Ok(())
   }
+  pub async fn quad_flush_all(&mut self) -> Result<(),Error> {
+    let qs = self.quad_updates.read().await.keys().copied().collect::<Vec<QuadId>>();
+    self.quad_flush_by_ids(&qs).await
+  }
   pub async fn quad_flush_partial(&mut self, n: usize) -> Result<(),Error> {
     let mut qs = self.quad_updates.read().await.keys()
-      .filter(|q_id| { self.quad_count.get(q_id).copied().unwrap_or(0) > 0 })
+      .filter(|q_id| { self.quad_get_update_count(q_id) > 0 })
       .copied()
       .collect::<Vec<QuadId>>();
     qs.sort_unstable_by(|a,b| {
-      let ca = self.quad_count.get(a).copied().unwrap_or(0);
-      let cb = self.quad_count.get(b).copied().unwrap_or(0);
+      let ca = self.quad_get_update_count(a);
+      let cb = self.quad_get_update_count(b);
       cb.cmp(&ca) // descending
     });
-    for q_id in &qs[0..n.min(qs.len())] {
-      //println!["flush {} (top: {})", q_id, self.quad_count.get(q_id).copied().unwrap_or(0)];
-      self.quad_flush_by_id(*q_id).await?;
-    }
+    self.quad_flush_by_ids(&qs[0..n.min(qs.len())]).await?;
     let mut queue = vec![];
     for q_id in self.quad_updates.read().await.keys() {
       let age = {
@@ -393,10 +447,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         queue.push(*q_id);
       }
     }
-    for q_id in queue {
-      self.quad_flush_by_id(q_id).await?;
-    }
-    Ok(())
+    self.quad_flush_by_ids(&queue).await
   }
 
   pub async fn id_flush_by_block(&mut self, b: IdBlock) -> Result<(),Error> {
@@ -404,54 +455,74 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     if o_ids.is_none() { return Ok(()) }
     let ids = o_ids.unwrap();
     let ifile = id_file_from_block(b);
-    let mut s = self.open_file_rw(&ifile).await?;
-    let buf = {
-      if let Some(xids) = self.id_cache.get_mut(&b) {
-        for (id,q_id) in ids.iter() {
-          xids.insert(*id, *q_id);
-        }
-        pack_ids(&xids)
-      } else {
-        let mut buf = vec![];
-        s.read_to_end(&mut buf).await?;
-        s.seek(io::SeekFrom::Start(0)).await?;
-        let mut xids = unpack_ids(&buf)?;
-        for (id,q_id) in ids.iter() {
-          xids.insert(*id,*q_id);
-        }
-        pack_ids(&xids)
-      }
-    };
-    s.set_len(buf.len() as u64).await?;
+    let mut s = self.open_file_a(&ifile).await?;
+    let buf = pack_ids(&ids);
     s.write_all(&buf).await?;
     s.flush().await?;
-    self.id_update_count -= ids.len() as u64;
-    self.id_cache.put(b,ids);
-    self.id_count.insert(b, 0);
-    self.close_file(&ifile);
+    self.id_update_count_total -= ids.len() as u64;
+    {
+      let mut ic = self.id_cache.lock().await;
+      for (id,q_id) in ids {
+        ic.put(id,q_id);
+      }
+    }
+    self.id_update_count.insert(b, 0);
+    self.id_update_age.insert(b, 0);
+    self.close_file(&ifile).await;
+    Ok(())
+  }
+  pub async fn id_flush_by_blocks(&mut self, bs: &[IdBlock]) -> Result<(),Error> {
+    let mut work: Vec<task::JoinHandle<
+      Result<(IdBlock,HashMap<RecordId,QuadId>),Error>
+    >> = Vec::with_capacity(bs.len());
+    for br in bs {
+      let b = *br;
+      let iu = self.id_updates.clone();
+      let active_files = self.active_files.clone();
+      let storage = self.storage.clone();
+      work.push(task::spawn(async move {
+        let o_ids = iu.write().await.remove(&b);
+        if o_ids.is_none() { return Ok((b,HashMap::new())) }
+        let ids = o_ids.unwrap();
+        let ifile = id_file_from_block(b);
+        let mut s = Self::open_file_a_params(&ifile, active_files.clone(), storage).await?;
+        let buf = pack_ids(&ids);
+        s.write_all(&buf).await?;
+        s.flush().await?;
+        Ok((b,ids))
+      }));
+    }
+    for r in futures::future::join_all(work).await {
+      let (b,ids) = r?;
+      self.id_update_count_total -= ids.len() as u64;
+      {
+        let mut ic = self.id_cache.lock().await;
+        for (id,q_id) in ids {
+          ic.put(id,q_id);
+        }
+      }
+      self.id_update_count.insert(b, 0);
+      self.id_update_age.insert(b, 0);
+      let ifile = id_file_from_block(b);
+      self.close_file(&ifile).await;
+    }
     Ok(())
   }
   pub async fn id_flush_all(&mut self) -> Result<(),Error> {
     let blocks = self.id_updates.read().await.keys().copied().collect::<Vec<IdBlock>>();
-    for b in blocks {
-      self.id_flush_by_block(b).await?;
-    }
-    Ok(())
+    self.id_flush_by_blocks(&blocks).await
   }
   pub async fn id_flush_partial(&mut self, n: usize) -> Result<(),Error> {
     let mut qs = self.id_updates.read().await.keys()
-      .filter(|b| { self.id_count.get(b).copied().unwrap_or(0) > 0 })
+      .filter(|b| { self.id_update_count.get(b).copied().unwrap_or(0) > 0 })
       .copied()
       .collect::<Vec<QuadId>>();
     qs.sort_unstable_by(|a,b| {
-      let ca = self.id_count.get(a).copied().unwrap_or(0);
-      let cb = self.id_count.get(b).copied().unwrap_or(0);
+      let ca = self.id_update_count.get(a).copied().unwrap_or(0);
+      let cb = self.id_update_count.get(b).copied().unwrap_or(0);
       cb.cmp(&ca) // descending
     });
-    for b in &qs[0..n.min(qs.len())] {
-      //println!["flush {} (top: {})", q_id, self.quad_count.get(q_id).copied().unwrap_or(0)];
-      self.id_flush_by_block(*b).await?;
-    }
+    self.id_flush_by_blocks(&qs[0..n.min(qs.len())]).await?;
     let mut queue = vec![];
     for b in self.id_updates.read().await.keys() {
       let age = {
@@ -467,10 +538,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         queue.push(*b);
       }
     }
-    for b in queue {
-      self.id_flush_by_block(b).await?;
-    }
-    Ok(())
+    self.id_flush_by_blocks(&queue).await
   }
 
   pub async fn missing_flush(&mut self) -> Result<(),Error> {
@@ -484,7 +552,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     s.write_all(&buf).await?;
     s.flush().await?;
     self.missing_updates.clear();
-    self.close_file(&mfile);
+    self.close_file(&mfile).await;
     Ok(())
   }
   pub async fn flush(&mut self) -> Result<(),Error> {
@@ -494,42 +562,123 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     self.missing_flush().await?;
     Ok(())
   }
-  pub async fn get_record(&mut self, id: RecordId) -> Result<Option<R>,Error> {
+  async fn get_qid_for_id(&mut self, id: RecordId) -> Result<Option<QuadId>,Error> {
     let b = self.id_block(id);
     let mut o_q_id = self.id_updates.read().await.get(&b).and_then(|ids| ids.get(&id).copied());
     if o_q_id.is_none() {
-      o_q_id = self.id_cache.get(&b).and_then(|ids| ids.get(&id)).copied();
+      o_q_id = self.id_cache.lock().await.get(&id).copied();
     }
-    let q_id = if let Some(q_id) = o_q_id { q_id } else {
-      let ifile = self.id_file(id);
-      let mut buf = vec![];
-      if let Some(mut s) = self.open_file_r(&ifile).await? {
-        s.read_to_end(&mut buf).await?;
+    if let Some(q_id) = o_q_id {
+      return Ok(Some(q_id));
+    }
+    let ifile = self.id_file(id);
+    let mut buf = vec![];
+    if let Some(mut s) = self.open_file_r(&ifile).await? {
+      s.read_to_end(&mut buf).await?;
+    }
+    let mut ids = HashMap::new();
+    {
+      let mut offset = 0;
+      while offset < buf.len() {
+        let s = unpack_ids(&buf[offset..], &mut ids)?;
+        offset += s;
+        if s == 0 { break }
       }
-      let ids = unpack_ids(&buf)?;
-      let g = ids.get(&id).copied();
-      self.id_cache.put(b, ids);
-      self.close_file(&ifile);
-      if g.is_none() { return Ok(None) }
-      g.unwrap()
-    };
+    }
+    let g = ids.get(&id).copied();
+    {
+      let mut ic = self.id_cache.lock().await;
+      for (r_id,q_id) in ids {
+        ic.put(r_id, q_id);
+      }
+    }
+    self.close_file(&ifile).await;
+    Ok(g)
+  }
+  async fn get_qid_for_ids(&mut self, ids: &[RecordId]) -> Result<HashMap<RecordId,QuadId>,Error> {
+    let mut work: Vec<task::JoinHandle<
+      Result<Option<(RecordId,QuadId)>,Error>
+    >> = Vec::with_capacity(ids.len());
+    for xid in ids {
+      let id = *xid;
+      let b = self.id_block(id);
+      let ic = self.id_cache.clone();
+      let iu = self.id_updates.clone();
+      let ifile = self.id_file(id);
+      let active_files = self.active_files.clone();
+      let storage = self.storage.clone();
+      work.push(task::spawn(async move {
+        let mut o_q_id = iu.read().await.get(&b).and_then(|ids| ids.get(&id).copied());
+        if o_q_id.is_none() {
+          o_q_id = ic.lock().await.get(&id).copied();
+        }
+        if let Some(q_id) = o_q_id {
+          return Ok(Some((id,q_id)));
+        }
+        let mut buf = vec![];
+        if let Some(mut s) = Self::open_file_r_params(&ifile, active_files.clone(), storage).await? {
+          s.read_to_end(&mut buf).await?;
+        }
+        let mut ids = HashMap::new();
+        {
+          let mut offset = 0;
+          while offset < buf.len() {
+            let s = unpack_ids(&buf[offset..], &mut ids)?;
+            offset += s;
+            if s == 0 { break }
+          }
+        }
+        let g = ids.get(&id).copied();
+        {
+          let mut icl = ic.lock().await;
+          for (r_id,q_id) in ids {
+            icl.put(r_id, q_id);
+          }
+        }
+        Self::close_file_params(&ifile, active_files).await;
+        Ok(g.map(|q_id| (id,q_id)))
+      }));
+    }
+    let mut results = HashMap::new();
+    for r in futures::future::join_all(work).await {
+      if let Some((r_id,q_id)) = r? {
+        results.insert(r_id, q_id);
+      }
+    }
+    Ok(results)
+  }
+  pub async fn get_record(&mut self, id: RecordId) -> Result<Option<R>,Error> {
+    if let Some(record) = self.record_cache.get(&id).cloned() {
+      return Ok(Some(record));
+    }
+    let o_q_id = self.get_qid_for_id(id).await?;
+    if o_q_id.is_none() { return Ok(None) }
+    let q_id = o_q_id.unwrap();
     if let Some(records) = self.quad_updates.read().await.get(&q_id) {
       if let Some(r) = records.as_ref().and_then(|items| items.get(&id).cloned()) {
         return Ok(Some(r));
       }
     }
-    if let Some(records) = self.quad_cache.get(&q_id) {
-      return Ok(records.get(&id).cloned());
-    }
     let qfile = quad_file(q_id);
     let mut buf = vec![];
     if let Some(mut s) = self.open_file_r(&qfile).await? {
       s.read_to_end(&mut buf).await?;
     }
-    let records = R::unpack(&buf)?;
+    let records = {
+      let mut offset = 0;
+      let mut records = HashMap::new();
+      while offset < buf.len() {
+        let s = R::unpack(&buf[offset..], &mut records)?;
+        offset += s;
+        if s == 0 { break }
+      }
+      records
+    };
     let r = records.get(&id).cloned();
-    self.quad_cache.put(q_id, records);
-    self.close_file(&qfile);
+    for (r_id,r) in records {
+      self.record_cache.put(r_id, r);
+    }
+    self.close_file(&qfile).await;
     Ok(r)
   }
   async fn get_position(&mut self, record: &R) -> Result<Option<Position>,Error> {
@@ -547,7 +696,8 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     let record = o_r.unwrap();
     Ok(record.get_position())
   }
-  pub async fn split_quad(&mut self, q_id: &QuadId, bbox: &BBox) -> Result<(),Error> {
+  pub async fn split_quad(&mut self, q_id: &QuadId) -> Result<(),Error> {
+    let bbox = self.quad_bbox.get(q_id).copied().unwrap();
     let qfile = quad_file(*q_id);
     let records = self.read_quad(*q_id).await?;
     {
@@ -558,6 +708,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     }
     self.lock_file(&qfile).await;
     self.quad_count.insert(*q_id, 0);
+    self.quad_update_count.insert(*q_id, 0);
     let (nx,ny) = (2,2);
     let mut quads = vec![];
     for i in 0..nx {
@@ -599,10 +750,13 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         if q.1.is_empty() {
           self.quad_updates.write().await.insert(*q_id, None);
           self.quad_count.insert(*q_id, 0);
+          self.quad_update_count.insert(*q_id, 0);
         } else {
           self.quad_count.insert(*q_id, q.1.len() as u64);
+          self.quad_update_count.insert(*q_id, q.1.len() as u64);
           self.quad_updates.write().await.insert(*q_id, Some(q.1));
         }
+        self.quad_bbox.insert(*q_id, q.0.clone());
         nchildren.push(QTree::Quad { id: *q_id, bbox: q.0.clone() });
       } else {
         let id = self.next_quad_id;
@@ -613,10 +767,13 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         if q.1.is_empty() {
           self.quad_updates.write().await.insert(id, None);
           self.quad_count.insert(id, 0);
+          self.quad_update_count.insert(id, 0);
         } else {
           self.quad_count.insert(id, q.1.len() as u64);
+          self.quad_update_count.insert(id, q.1.len() as u64);
           self.quad_updates.write().await.insert(id, Some(q.1));
         }
+        self.quad_bbox.insert(id, q.0.clone());
         nchildren.push(QTree::Quad { id, bbox: q.0.clone() });
       }
       i += 1;
@@ -633,7 +790,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
           match c {
             QTree::Node { children, .. } => {
               ncursors.extend(children.iter_mut()
-                .filter(|ch| { bbox_overlap(bbox,&ch.bbox()) })
+                .filter(|ch| { bbox_overlap(&bbox,&ch.bbox()) })
                 .collect::<Vec<_>>());
             },
             QTree::Quad { id, .. } => {
@@ -650,29 +807,62 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
       }
       assert![found, "did not locate quad id={}", q_id];
     }
-    self.close_file(&qfile);
+    self.close_file(&qfile).await;
     self.check_flush().await?;
     Ok(())
   }
-  pub async fn get_quads(&mut self, records: &[R]) -> Result<HashMap<QuadId,(BBox,Vec<usize>)>,Error> {
-    let mut result: HashMap<QuadId,(BBox,Vec<usize>)> = HashMap::new();
+  async fn get_quads(&mut self, records: &[R]) -> Result<HashMap<QuadId,Vec<usize>>,Error> {
+    let mut result: HashMap<QuadId,Vec<usize>> = HashMap::new();
     let mut positions = HashMap::new();
+
     let mut rmap = HashMap::new();
     for r in records.iter() {
       rmap.insert(r.get_id(), r.clone());
     }
-    for (i,r) in records.iter().enumerate() {
-      let mut o_p = check_position_records(r, &rmap);
-      if o_p.is_none() {
-        o_p = self.get_position(r).await?;
-      }
-      if let Some(p) = o_p {
-        positions.insert(i,p);
-      } else {
-        self.missing_updates.insert(r.get_id(),r.clone());
-        self.missing_count += 1;
+
+    let mut get_qids = vec![];
+    {
+      let mut ic = self.id_cache.lock().await;
+      let iu = self.id_updates.read().await;
+      for (i,r) in records.iter().enumerate() {
+        let id = r.get_id();
+        let b = self.id_block(id);
+        let mut o_q_id = iu.get(&b).and_then(|ids| ids.get(&id).copied());
+        if o_q_id.is_none() {
+          o_q_id = ic.get(&id).copied();
+        }
+        if let Some(q_id) = o_q_id {
+          if let Some(items) = result.get_mut(&q_id) {
+            items.push(i);
+          } else {
+            result.insert(q_id, vec![i]);
+          }
+        } else if let Some(p) = check_position_records(r, &rmap) {
+          positions.insert(i,p);
+        } else if let Some(f_id) = r.get_refs().first() {
+          get_qids.push(*f_id);
+        }
       }
     }
+    {
+      let qids = self.get_qid_for_ids(&get_qids).await?;
+      for (i,r) in records.iter().enumerate() {
+        if positions.contains_key(&i) { continue }
+        if let Some(f_id) = r.get_refs().first() {
+          if let Some(q_id) = qids.get(&f_id) {
+            if let Some(items) = result.get_mut(&q_id) {
+              items.push(i);
+            } else {
+              result.insert(*q_id, vec![i]);
+            }
+          } else {
+            self.missing_updates.insert(r.get_id(),r.clone());
+            self.missing_count += 1;
+          }
+        }
+      }
+    }
+
     let mut cursors = vec![&self.root];
     let mut ncursors = vec![];
     while !cursors.is_empty() {
@@ -688,10 +878,10 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
           QTree::Quad { id, bbox } => {
             positions.drain_filter(|i,p| {
               if overlap(p,bbox) {
-                if let Some((_,items)) = result.get_mut(id) {
+                if let Some(items) = result.get_mut(id) {
                   items.push(*i);
                 } else {
-                  result.insert(*id, (bbox.clone(),vec![*i]));
+                  result.insert(*id, vec![*i]);
                 }
                 true
               } else {
@@ -709,38 +899,73 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     Ok(result)
   }
   async fn lock_file(&mut self, file: &String) {
-    if let Some(active) = self.active_files.get(file) {
+    if let Some(active) = self.active_files.read().await.get(file) {
       active.lock().await;
     }
     let active = Arc::new(Mutex::new(()));
     let ac = active.clone();
-    self.active_files.insert(file.clone(), active);
+    self.active_files.write().await.insert(file.clone(), active);
     ac.lock().await;
   }
   async fn open_file_r(&mut self, file: &String) -> Result<Option<S>,Error> {
-    if let Some(active) = self.active_files.get(file) {
+    Self::open_file_r_params(file, self.active_files.clone(), self.storage.clone()).await
+  }
+  async fn open_file_r_params(
+    file: &String,
+    active_files: Arc<RwLock<HashMap<String,Arc<Mutex<()>>>>>,
+    storage: Arc<Mutex<Box<dyn Storage<S>>>>,
+  ) -> Result<Option<S>,Error> {
+    if let Some(active) = active_files.read().await.get(file) {
       active.lock().await;
     }
-    let mut st = self.storage.lock().await;
+    let mut st = storage.lock().await;
     let active = Arc::new(Mutex::new(()));
     let ac = active.clone();
-    self.active_files.insert(file.clone(), active);
+    active_files.write().await.insert(file.clone(), active);
     ac.lock().await;
     st.open_r(file).await
   }
   async fn open_file_rw(&mut self, file: &String) -> Result<S,Error> {
-    if let Some(active) = self.active_files.get(file) {
+    if let Some(active) = self.active_files.read().await.get(file) {
       active.lock().await;
     }
     let mut st = self.storage.lock().await;
     let active = Arc::new(Mutex::new(()));
     let ac = active.clone();
-    self.active_files.insert(file.clone(), active);
+    self.active_files.write().await.insert(file.clone(), active);
     ac.lock().await;
     st.open_rw(file).await
   }
-  fn close_file(&mut self, file: &String) {
-    self.active_files.remove(file);
+  async fn open_file_a(&mut self, file: &String) -> Result<S,Error> {
+    Self::open_file_a_params(
+      file,
+      self.active_files.clone(),
+      self.storage.clone()
+    ).await
+  }
+  async fn open_file_a_params(
+    file: &String,
+    active_files: Arc<RwLock<HashMap<String,Arc<Mutex<()>>>>>,
+    storage: Arc<Mutex<Box<dyn Storage<S>>>>,
+  ) -> Result<S,Error> {
+    if let Some(active) = active_files.read().await.get(file) {
+      active.lock().await;
+    }
+    let mut st = storage.lock().await;
+    let active = Arc::new(Mutex::new(()));
+    let ac = active.clone();
+    active_files.write().await.insert(file.clone(), active);
+    ac.lock().await;
+    st.open_a(file).await
+  }
+  async fn close_file(&mut self, file: &String) {
+    Self::close_file_params(file, self.active_files.clone()).await
+  }
+  async fn close_file_params(
+    file: &String,
+    active_files: Arc<RwLock<HashMap<String,Arc<Mutex<()>>>>>,
+  ) {
+    active_files.write().await.remove(file);
   }
   pub async fn finish(&mut self) -> Result<(),Error> {
     let mut prev_count = 0;
@@ -755,9 +980,17 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
           if let Some(mut s) = self.open_file_r(&mfile).await? {
             s.read_to_end(&mut buf).await?;
           }
-          self.close_file(&mfile);
+          self.close_file(&mfile).await;
+        }
+        let mut records = HashMap::new();
+        {
+          let mut offset = 0;
+          while offset < buf.len() {
+            let s = R::unpack(&buf, &mut records)?;
+            offset += s;
+            if s == 0 { break }
+          }
         }
-        let mut records = R::unpack(&buf)?;
         self.add_records(&records.drain().map(|(_,r)| r).collect::<Vec<R>>()).await?;
       }
       self.add_records(&m_records).await?;
@@ -793,7 +1026,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     let buf = self.get_meta().to_bytes()?;
     s.set_len(buf.len() as u64).await?;
     s.write_all(&buf).await?;
-    self.close_file(&mfile);
+    self.close_file(&mfile).await;
     Ok(())
   }
   fn id_block(&self, id: RecordId) -> IdBlock {
@@ -826,31 +1059,26 @@ fn missing_file(m_id: u64) -> String {
 fn id_file_from_block(b: IdBlock) -> String {
   format!["i/{:02x}/{:x}",b%256,b/256]
 }
-fn unpack_ids(buf: &[u8]) -> Result<HashMap<RecordId,QuadId>,Error> {
-  let mut records = HashMap::new();
-  if buf.is_empty() { return Ok(records) }
+fn unpack_ids(buf: &[u8], records: &mut HashMap<RecordId,QuadId>) -> Result<usize,Error> {
+  if buf.is_empty() { return Ok(0) }
   let mut offset = 0;
-  let (s,len) = varint::decode(&buf[offset..])?;
-  offset += s;
-  for _ in 0..len {
+  while offset < buf.len() {
     let (s,r_id) = varint::decode(&buf[offset..])?;
     offset += s;
     let (s,q_id) = varint::decode(&buf[offset..])?;
     offset += s;
     records.insert(r_id, q_id);
   }
-  Ok(records)
+  Ok(offset)
 }
 fn pack_ids(records: &HashMap<RecordId,QuadId>) -> Vec<u8> {
   let mut size = 0;
-  size += varint::length(records.len() as u64);
   for (r_id,q_id) in records {
     size += varint::length(*r_id);
     size += varint::length(*q_id);
   }
   let mut buf = vec![0;size];
   let mut offset = 0;
-  offset += varint::encode(records.len() as u64, &mut buf[offset..]).unwrap();
   for (r_id,q_id) in records {
     offset += varint::encode(*r_id, &mut buf[offset..]).unwrap();
     offset += varint::encode(*q_id, &mut buf[offset..]).unwrap();
